---------------------------------------------------------------------------------------
--------------------------------------- SCROLLER --------------------------------------
---------------------------------- version 2021.09.02 ---------------------------------

--[[ USAGE
		1 - create a frame us usual
		2 - fill the scroll properties table function: frame height can be resizable by defining minHeight and maxHeight properties
		3 - fill the column properties table (at least one column is required)
		4 - insert .scrollCreate(yourScrollFrame, scrollProperties, columnsProperties [, dataToFill]) -- creates structure
		5 - define what to do when scroll updates with .scrollLineUpdate(scrollFrame, lineIndex, dataLineIndex) function (test scrollFrame value if you have several scroll frames).
		6 - if mouse scripts were assigned to colums properties, write these scripts
		7 - when needed in the addon, call ".scrollUpdate(yourScrollFrame, dataToFill)" function to update its content
		8 - when needed to modify frame's height (if minHeight and maxHeight properties have been defined), call ".setHeight(yourScrollFrameOrName, newHeight)". When called, yourScrollFrame.height property is updated.
		
		Columns' properties will override general scroll properties
		
		Utilities functions:
			.scrollGetDataLine(scrollFrameOrName, lineIndex)
				-- returns:
				-- 		original data's line index
				--		original data's line data
			.scrollGetProperties(scrollFrameOrName): returns effective scroll properties
			.scrollPrintProperties(scrollFrameOrName): prints effective scroll properties.
			.scrollGetColumnNumber(scrollFrameOrName): returns the number of columns
			.scrollGetColumnProperties(scrollFrameOrName, columnIndex): returns effective column properties
			.scrollPrintColumnProperties(scrollFrameOrName, columnIndex): prints effective column properties
		
		[yourFrameName].[propertyName] -- scroll property
			[yourFrameName].currentLines -- number of lines currently visible
		[yourFrameName].column[i].[propertyName] -- column [i] property
		
		[yourFrameName].line[i] -- many informations, including dataLineIndex, data's line displayed
		[yourFrameName].bkLine[i] -- many informations, including dataLineIndex, data's line displayed
		[yourFrameName].line[i].texture
		
		[yourFrameName].line[i].[column].[i].text
		[yourFrameName].line[i].[column].[i].texture
		
		[yourFrameName].data -- all dataToFill
		
		Frame lines name are: [yourFrameName]_line_[lineNumber]
		Frame columns name are: [yourFrameName]_line_[lineNumber]_[columnNumber]
--]]

------------------------------------------------------------------- SCROLL FRAME DEFAULT GENERAL PROPERTIES
function Emlib.scroll_getDefaultProperties()
	local tbl = {
		-------------------- height: at least one of these properties must be set
		["height"] = 240,
		["minHeight"] = 0, -- optional: must be set if frame's height can be resized (min = 60 px)
		["maxHeight"] = 0, -- optional: must be set if frame's height can be resized
		["resizeButton"] = true, -- optional: if true, draw a resizing button at the frame's bottom
		["resizeScript"] = "Emlib.scrollFrameResized", -- only needed if frame's height can be resized and some other UI elements have to be updated (NIL for no script)
		--------------------
		["width"] = 200,
		["marginH"] = 4, -- px : horizontal margin from background frame
		["marginV"] = 4, -- px : vertical margin from background frame
		["alwaysShowArrow"] = true, -- if true, always show scroller up and down arrows
		-------------------- lines
		["lineHeight"] = 16, -- px
		["lineSpacing"] = 1, -- px
		["cellSpacing"] = 0, -- px // horizontal spacing, if several columns in a line
		-------------------- lines font
		["font"] = "arial", -- "friz" | "arial" | pathToLocalFont accepted
		["fontSize"] = 12,
		["fontOutline"] = false,
		["fontColor"] = "white", -- Hexacolor | RGB table | color name
		["fontAlign"] = "LEFT", -- "LEFT" | "RIGHT" | "CENTER"
		["fontAlpha"] = 1,
		["padding"] = nil, -- integer, or fontSize/3 by default
		-------------------- lines background
		["background"] = {0.1, 0.1, 0.1}, -- color, or texture file path
		["alpha"] = 1,
		["border"] = 0, -- border size, only available if ["background"] is not a texture path
		["borderColor"] = {0.5, 0.5, 0.5}, -- only available if ["background"] is not a texture path
		["mouseOverLineColor"] = nil, --{0.2, 0.2, 0.1}, -- color for visual effect on mouseOver the lines (NIL for no mouseOver effect)
		-------------------- scroller buttons images
		["defaultScroller"] = false, -- set to true to standard WoW scroller
		["scrollerWidth"] = 16, -- px : unused if defaultScroller set to true
		["scrollerColor"] = "111122", -- unused if defaultScroller set to true
		["scrollerBarAlpha"] = 0.6, -- unused if defaultScroller set to true
	}
	return tbl
end

------------------------------------------------------------------- SUB-LINE (COLUMNS) DEFAULT DEFINITION
function Emlib.scroll_getColumnDefaultProperties()
	local tbl = {
		---------- required
		["width"] = 0, -- % value accepted. if 0 or NIL (only for last column): all the remaining line's width
		---------- optional
		["template"] = nil, -- template name, as defined in XML file
		["templateType"] = nil, -- template type, as defined in XML file : "StatusBar" | "EditBox"...
		["type"] = "Frame", -- "Frame" | "Button" | "EditBox" (NIL for default frame)
		["multilines"] = false, -- optional value for "EditBox" type
		["height"] = nil, -- for multilines "EditBox"
		---------- script (only needed for button)
		["script"] = "Emlib.scrollFrameClickLineScript_", -- NIL for no script
		---------- optional font properties (override scroller properties)
		["font"] = nil,
		["fontSize"] = nil,
		["fontOutline"] = nil,
		["fontColor"] = nil,
		["fontAlign"] = nil,
		["fontAlpha"] = nil,
		["padding"] = nil,
		---------- optional background properties (override scroller properties)
		["background"] = nil,
		["alpha"] = nil,
		["border"] = nil,
		["borderColor"] = nil,
		["mouseOverLineColor"] = nil,
	}
	return tbl
end

------------------------------------------------------------------- UTILITIES

function Emlib.setHeight(scrollFrameOrName, newHeight) -- utility for resizable scroll frame
	local newHeight = Emlib.strToNumber(newHeight) -- 0 if nil
	local frame = Emlib.scroll_getRefFrame(scrollFrameOrName, false)
	if (frame ~= nil) then
		if (frame.resizable) then
			local effectiveHeight = min(frame.maxHeight, max(newHeight, frame.minHeight))
			-------------- adapt to line height
			local numLines = Emlib.round(effectiveHeight / frame.lineHeight, 0)
			frame.height = (numLines * frame.lineHeight) + (frame.marginV * 2)
--print(".setHeight()---newHeight=", newHeight, "---numLines=", numLines, "---effHeight=", effectiveHeight, "---Hproperty", frame.height)--/////
			frame:SetHeight(frame.height)
			frame.scroller:SetHeight(frame.height)
			Emlib.scrollUpdate(frame)
		end
	end
end

function Emlib.scrollGetDataLine(scrollFrameOrName, lineIndex) -- utility
	-- returns:
	-- 		original data's line index
	--		original data's line data
	
	local lineIndex = lineIndex or 0
	local frame = Emlib.scroll_getRefFrame(scrollFrameOrName, false)
	if (frame ~= nil) then
		if (lineIndex > 0 and lineIndex <= Emlib.scroll_getNumberOfLinesToDisplay(frame)) then
			local numLi = frame.line[lineIndex].dataLineIndex -- line number of original data
			return numLi, frame.data[numLi]
		else
			return lineIndex, nil -- line lineIndex is not displayed in scroll frame
		end
	else
		return nil
	end
end

function Emlib.scrollGetColumnNumber(scrollFrameOrName) -- utility
	-- returns number of columns
	
	local frame = Emlib.scroll_getRefFrame(scrollFrameOrName, false)
	if (frame ~= nil) then
		return frame.columns
	else
		return 0
	end
end

------------------------------------ properties utilities

function Emlib.scrollGetProperties(scrollFrameOrName) -- utility: returns effective scroll properties
	local frameName = scrollFrameOrName
	if (type(scrollFrameOrName) == "string") then -- a frame name was passed
		scrollFrameOrName = _G[scrollFrameOrName]
	else
		frameName = scrollFrameOrName:GetName()
	end
	
	local tbl = {}
	local tblPropsList = Emlib.scroll_getDefaultProperties()
	local propVal = ""
	for propName, value in pairs(tblPropsList) do
		propVal = scrollFrameOrName[propName]
		if (type(propVal) == "table") then
			propVal = Emlib.scroll_dumpTableOne(propVal)
		elseif (type(propVal) == "boolean") then
			propVal = Emlib.booleanToStr(propVal)
		end
		tbl[propName] = Emlib.nilSiNul(propVal)
	end
	tbl["scroller"] = Emlib.nilSiNul(scrollFrameOrName.scroller:GetName())
	tbl["slider"] = Emlib.nilSiNul(scrollFrameOrName.slider:GetName())
	return tbl
end

function Emlib.scrollPrintProperties(scrollFrameOrName) -- utility: prints effective scroll properties
	local frameName = scrollFrameOrName
	if (type(scrollFrameOrName) == "string") then -- a frame name was passed
		scrollFrameOrName = _G[scrollFrameOrName]
	else
		frameName = scrollFrameOrName:GetName()
	end
	
	Emlib.echoChatAppl(frameName.." properties", false, "yellow")
	Emlib.echoDebug(Emlib.scrollGetProperties(scrollFrameOrName, columnIndex), "-------------------------------")
end

function Emlib.scrollGetColumnProperties(scrollFrameOrName, columnIndex) -- utility: returns effective column properties
	local frameName, frame
	local columnIndex = tonumber(columnIndex) or 0
	local frame = Emlib.scroll_getRefFrame(scrollFrameOrName, false)
	if (frame ~= nil) then return (frame.column[columnIndex]) else return nil end
end

function Emlib.scrollPrintColumnProperties(scrollFrameOrName, columnIndex) -- utility: prints effective column properties
	local frameName, frame
	if (type(scrollFrameOrName) ~= "string") then
		frameName = scrollFrameOrName:GetName()
		frame = scrollFrameOrName
	else
		frameName = scrollFrameOrName
		frame = _G[scrollFrameOrName]
	end
	
	Emlib.echoChatAppl(frameName.." column "..columnIndex.." properties", false, "yellow")
	Emlib.echoDebug(Emlib.scrollGetColumnProperties(frame, columnIndex), "-------------------------------")
end

------------------------------------------------------------------- INTERNAL UTILITIES

function Emlib.scroll_getNumberOfLinesToDisplay(scrollFrame) -- utility (internal)
--print(".scroll_getNumberOfLinesToDisplay()----scrollFrame=", scrollFrame:GetName(), "---height=", scrollFrame:GetHeight(), "---lineHeight=", scrollFrame.lineHeight, "---ret=", floor(scrollFrame:GetHeight() / scrollFrame.lineHeight))--/////
	scrollFrame.currentLines = floor(Emlib.round(scrollFrame:GetHeight(), 0, false) / scrollFrame.lineHeight)
	return scrollFrame.currentLines
end

function Emlib.scroll_getFrameLine(scrollFrame, lineIndex) -- utility (internal)
	-- returns a reference to the passed line frame
	return scrollFrame.line[lineIndex]
end

function Emlib.scroll_getColumn(scrollFrame, lineIndex, columnIndex) -- utility (internal)
	return scrollFrame.line[lineIndex].column[columnIndex]
end

function Emlib.scroll_dumpTableOne(tbl) -- mini-dump: only one level !
	local strResult = ""
	for key, value in pairs(tbl) do
		strResult = strResult.."   ["..key.."]"..value
	end
	return strResult
end

function Emlib.scroll_getRefFrame(scrollFrameOrName, frameNameBoolean) -- utility: returns a frame or its name, or nil
	local frameNameBoolean = Emlib.isTrue(frameNameBoolean)
	local frameName, frame
	if (scrollFrameOrName ~= nil) then
		if (type(scrollFrameOrName) ~= "string") then
			frameName = scrollFrameOrName:GetName()
			frame = scrollFrameOrName
		else
			frameName = scrollFrameOrName
			frame = _G[scrollFrameOrName]
		end
		if (frameNameBoolean) then return frameName else return frame end
	else
		return nil
	end
end

function Emlib.scroll_checkProp(userProps, defaultProps, propName, propType) -- general scroll
	-- type : "nil", "number", "string", "boolean", "table", "function", "thread", or "userdata"
	
	if (userProps[propName] == nil) then
		------- replace by default
		userProps[propName] = defaultProps[propName]
	else
		------- passed: check type
		if (propType ~= nil) then -- specified prop type to be checked
			userProps[propName] = Emlib.scroll_checkTypeProp(propType, userProps[propName], defaultProps[propName])
			if (userProps[propName] == nil) then userProps[propName] = defaultProps[propName] end
		end
	end
	return userProps[propName]
end

function Emlib.scroll_checkColumnProp(userProps, defaultColumnProps, defaultScrollProps, propName, propType) -- column
	-- type : "nil", "number", "string", "boolean", "table", "function", "thread", or "userdata"
	if (userProps[propName] == nil) then
		if (defaultScrollProps[propName] == nil) then
			return defaultColumnProps[propName] -- replace by column default
		else
			return defaultScrollProps[propName] -- replace by scroll default
		end
	else
		if (propType ~= nil) then -- specified prop type to be checked
			userProps[propName] = Emlib.scroll_checkTypeProp(propType, userProps[propName], defaultScrollProps[propName])
			if (userProps[propName] == nil) then
				return Emlib.scroll_checkColumnProp(userProps, defaultColumnProps, defaultScrollProps, propName, propType)
			else
				return userProps[propName]
			end
		else
			return userProps[propName]
		end
	end
end

function Emlib.scroll_checkTypeProp(propType, userPropValue, defaultPropValue) -- return NIL if incorrect prop type
	if (propType == "boolean") then
		userPropValue = Emlib.isTrue(userPropValue)
	elseif (propType == "number") then
		userPropValue = tonumber(userPropValue) or -1
		if (userPropValue < 0) then userPropValue = nil end -- no negative number allowed
	else
		if (type(userPropValue) ~= propType) then userPropValue = nil end
	end
	return userPropValue
end

function Emlib.scroll_getProp(userProps, defaultProps, propName)
	if (userProps[propName] == nil) then
		return defaultProps[propName]
	else
		return userProps[propName]
	end
end

------------------------------------------------------------------- BACKGROUND FRAME FOR LIST AND SCROLLER

function Emlib.scrollCreate(frame, userScrollProps, userColumns, optionalDataTable)
	-- no needed data to create scroll frame: it will be filled on update
	local optionalDataTable = optionalDataTable or {}
	local defaultScrollProps = Emlib.scroll_getDefaultProperties()
	local scrollEffectiveProps
	frame, scrollEffectiveProps = Emlib.scroll_applyScrollerProps(frame, userScrollProps, defaultScrollProps) -- apply checked prefs to frame
	frame:SetSize(frame.width, frame.height + (frame.marginV * 2))
	frame.data = optionalDataTable -- needed for "OnVerticalScroll" script
	Emlib.scroll_createScroller(frame, userColumns, scrollEffectiveProps)
	------------
	local function setFrameLimits(frame, minWidth, maxWidth, minHeight, maxHeight) -- pour compat wow9/wow10
		frame:SetMinResize(tonumber(minWidth) or 10, tonumber(minHeight) or 10)--non wow10
		frame:SetMaxResize(tonumber(maxWidth) or 50, tonumber(maxHeight) or 50)--non wow10
	end
	frame:SetResizable(frame.resizable)
	if (frame.resizable) then
		local success, errorMessage = pcall(setFrameLimits, frame, userScrollProps.width, userScrollProps.width, userScrollProps.minHeight, userScrollProps.maxHeight)
		if (not success) then -- wow10
			frame:SetResizeBounds(userScrollProps.width, userScrollProps.minHeight, userScrollProps.width, userScrollProps.maxHeight)
		end
		if (frame.resizeButton) then Emlib.scroll_createResizer(frame) end
	end
	------------
	Emlib.scrollUpdate(frame, frame.data, true) -- force visible scroller on first update
end

function Emlib.scroll_createResizer(scrollFrame) -- btn resize @ bottom
	local resizeBtn = CreateFrame("Button", "$parent_resizeBtn", scrollFrame)
	resizeBtn:SetSize(64, 8)
	resizeBtn:SetPoint("TOP", scrollFrame, "BOTTOM", 0, -2)
	resizeBtn:SetNormalTexture(Emlib.getStuffPath("hamburger", true))
	resizeBtn:SetHighlightTexture(Emlib.getStuffPath("hamburger", true))
	resizeBtn:SetPushedTexture(Emlib.getStuffPath("hamburger", true))
	------------- new property
	local point, relativeTo, relativePoint, xOfs, yOfs = scrollFrame:GetPoint(1)
	scrollFrame.points = {
		["point"] = point,
		["relativeTo"] = relativeTo,
		["relativePoint"] = relativePoint,
		["xOfs"] = xOfs,
		["yOfs"] = yOfs,
	}
	------------- scripts
	resizeBtn:SetScript("OnMouseDown", function(self, mouse)
		if (mouse == "LeftButton" and not scrollFrame.isResizing) then
			scrollFrame.isResizing = true
			scrollFrame.scroller:Hide() -- visual effect (live update not possible ?)
			scrollFrame:StartSizing("BOTTOM")
		end
	end)
	resizeBtn:SetScript("OnMouseUp", function(self, mouse)
		if (mouse == "LeftButton" and scrollFrame.isResizing) then
			scrollFrame:StopMovingOrSizing()
			scrollFrame.isResizing = false
			scrollFrame.scroller:Show() -- visual effect
			--------- re-anchor nécessaire après un StartSizing !
			scrollFrame:ClearAllPoints()
			scrollFrame:SetPoint(scrollFrame.points.point, scrollFrame.points.relativeTo, scrollFrame.points.relativePoint, scrollFrame.points.xOfs, scrollFrame.points.yOfs)
			Emlib.setHeight(scrollFrame, scrollFrame:GetHeight())
			if (not Emlib.isEmpty(scrollFrame.resizeScript)) then RunScript(scrollFrame.resizeScript.."("..Emlib.kwote(scrollFrame:GetName())..","..Emlib.kwote(scrollFrame:GetHeight())..")") end
		end
	end)
end

----------------------------------------------------------------- SCROLLER UPDATE

function Emlib.scrollUpdate(scrollFrame, data, forceVisibleScrollerBoolean) -- updates all scroller's lines (and show or hides them)
	-- no need to pass data if data has not changed
	if (data == nil) then
		data = scrollFrame.data
	else
		if (type(data) ~= "table") then data = {data} end
		scrollFrame.data = data -- updated data
	end
	
	local totalDataLineNumber = #data
	local actualNumberOfLinesToDisplay = Emlib.scroll_getNumberOfLinesToDisplay(scrollFrame)
--print(".scrollUpdate(): A---", scrollFrame:GetName(), "--totalDataLineNumber=", totalDataLineNumber, "--actualNumberOfLinesToDisplay=", actualNumberOfLinesToDisplay, "--.maxFrameLines", scrollFrame.maxFrameLines)--/////
	local scroller = scrollFrame.scroller
	FauxScrollFrame_Update(scroller, totalDataLineNumber, actualNumberOfLinesToDisplay, scrollFrame.lineHeight, nil, nil, nil, nil, nil, nil, scrollFrame.alwaysShowArrow) -- set last arg to false to hide scroll buttons when not needed
	local invisLinesUp = floor(scroller.offsetPosUp/scrollFrame.lineHeight) -- NEW 9.x perso (voir Emlib.scroll_createScroller())
	local offset = 0
	for lineIndex = 1, scrollFrame.maxFrameLines do
		--local offset = lineIndex + FauxScrollFrame_GetOffset(scroller) -- line number of dataToDisplay
		offset = lineIndex + invisLinesUp -- NEW perso : line number of dataToDisplay

--print(".scrollUpdate(): B---", scroller:GetName(), "--max=", scrollFrame.maxFrameLines, "--lineIdx=", lineIndex, "--FSFOffset=", FauxScrollFrame_GetOffset(scroller), "--offset=", offset, "--AcliToDisplay=", actualNumberOfLinesToDisplay)--/////

		local frameLine = scrollFrame.bkLine[lineIndex]
		if (lineIndex > min(totalDataLineNumber, actualNumberOfLinesToDisplay)) then
			frameLine:Hide()
		else
			frameLine.dataLineIndex = offset -- needed for utility script .scrollGetDataLine()
			scrollFrame.line[lineIndex]["dataLineIndex"] = offset -- needed for utility script
			Emlib.scrollLineUpdate(scrollFrame, lineIndex, offset) -- user's define script (for each line)
			--------------------------------
			frameLine:Show() -- show the line
		end
	end
	
	if (Emlib.isTrue(forceVisibleScrollerBoolean) or scrollFrame.slider:GetValue() == 0) then scrollFrame.slider:SetValue(0) end -- pour bug affichage
end

----------------------------------------------------------------- INTERNAL

function Emlib.scroll_applyScrollerProps(frame, userScrollProps, defaultScrollProps)
	local tbl = {}
	-------------------------- replace default props by user props
	------------ special : height calculation
	tbl.height = Emlib.scroll_checkProp(userScrollProps, defaultScrollProps, "height", "number")
	tbl.minHeight = Emlib.scroll_checkProp(userScrollProps, defaultScrollProps, "minHeight", "number")
	tbl.maxHeight = Emlib.scroll_checkProp(userScrollProps, defaultScrollProps, "maxHeight", "number")
--print(".scroll_applyScrollerProps()---A---h=", tbl.height, "---min=", tbl.minHeight, "---max=", tbl.maxHeight)--////
	tbl.resizable = (not (tbl.minHeight == 0 or tbl.maxHeight == 0) and (tbl.minHeight ~= tbl.maxHeight))
	if (tbl.resizable) then
		if (tbl.minHeight > tbl.maxHeight) then
			local tmp = tbl.minHeight
			tbl.minHeight = tbl.maxHeight
			tbl.maxHeight = tmp
		end
		tbl.minHeight = max(60, tbl.minHeight) -- 60: graphic limit for scroller buttons
	else
		tbl.minHeight = tbl.height
		tbl.maxHeight = tbl.height
	end
--print(".scroll_applyScrollerProps()---B---h=", tbl.height, "---min=", tbl.minHeight, "---max=", tbl.maxHeight)--////
	------------ 
	tbl.width = Emlib.scroll_checkProp(userScrollProps, defaultScrollProps, "width", "number")
	tbl.marginV = Emlib.scroll_checkProp(userScrollProps, defaultScrollProps, "marginV", "number")
	tbl.marginH = Emlib.scroll_checkProp(userScrollProps, defaultScrollProps, "marginH", "number")
	tbl.alwaysShowArrow = Emlib.scroll_checkProp(userScrollProps, defaultScrollProps, "alwaysShowArrow", "boolean")
	
	tbl.lineHeight = Emlib.scroll_checkProp(userScrollProps, defaultScrollProps, "lineHeight", "number")
	tbl.lineSpacing = Emlib.scroll_checkProp(userScrollProps, defaultScrollProps, "lineSpacing", "number")
	tbl.cellSpacing = Emlib.scroll_checkProp(userScrollProps, defaultScrollProps, "cellSpacing", "number")
	tbl.font = Emlib.scroll_checkProp(userScrollProps, defaultScrollProps, "font", "string")
	tbl.fontSize = Emlib.scroll_checkProp(userScrollProps, defaultScrollProps, "fontSize", "number")
	
	local fontColor = Emlib.scroll_getProp(userScrollProps, defaultScrollProps, "fontColor")
	tbl.fontColor = Emlib.getColor(fontColor, "table")
	
	tbl.fontOutline = Emlib.scroll_checkProp(userScrollProps, defaultScrollProps, "fontOutline", "boolean")
	tbl.fontAlpha = Emlib.scroll_checkProp(userScrollProps, defaultScrollProps, "fontAlpha", "number")
	
	local fontAlign = strupper(Emlib.scroll_getProp(userScrollProps, defaultScrollProps, "fontAlign"))
	tbl.fontAlign = Emlib.scroll_getTextAlign(fontAlign, "LEFT")
	
	tbl.padding = Emlib.scroll_checkProp(userScrollProps, defaultScrollProps, "padding", "number")
		
	tbl.background = Emlib.scroll_getProp(userScrollProps, defaultScrollProps, "background")
	tbl.border = Emlib.scroll_checkProp(userScrollProps, defaultScrollProps, "border", "number")
	tbl.borderColor = Emlib.scroll_getProp(userScrollProps, defaultScrollProps, "borderColor")
	tbl.alpha = Emlib.scroll_checkProp(userScrollProps, defaultScrollProps, "alpha", "number")
	
	local mouseOverLineColor = Emlib.scroll_getProp(userScrollProps, defaultScrollProps, "mouseOverLineColor")
	tbl.mouseOverLineColor = mouseOverLineColor
	if (tbl.mouseOverLineColor ~= nil) then tbl.mouseOverLineColor = Emlib.getColor(mouseOverLineColor, "table") end
	
	-------------------------- scroller images and color
	tbl.defaultScroller = Emlib.scroll_checkProp(userScrollProps, defaultScrollProps, "defaultScroller", "boolean")
	tbl.scrollerColor = Emlib.scroll_getProp(userScrollProps, defaultScrollProps, "scrollerColor")
	tbl.scrollerColor = Emlib.getColor(tbl.scrollerColor, "table")
	tbl.scrollerWidth = Emlib.scroll_checkProp(userScrollProps, defaultScrollProps, "scrollerWidth", "number")
	tbl.scrollerBarAlpha = Emlib.scroll_getProp(userScrollProps, defaultScrollProps, "scrollerBarAlpha", "number")
	
	-------------------------- resize
	tbl.resizeButton = Emlib.scroll_checkProp(userScrollProps, defaultScrollProps, "resizeButton", "boolean")
	tbl.resizeScript = Emlib.scroll_getProp(userScrollProps, defaultScrollProps, "resizeScript")
	tbl.currentLines = 0 -- init
	
	-------------------------- apply properties
	for key, value in pairs(tbl) do
		frame[key] = value
	end
	
	-------------------------- 
	return frame, tbl -- tbl will be re-used for columns
end

function Emlib.scroll_getTextAlign(fontAlign, alignDefault) -- create FauxScrollFrame frame and lines
	local alignDefault = alignDefault or "LEFT"
	local tbl = {"LEFT", "RIGHT", "CENTER"}
	if (tContains(tbl, fontAlign)) then return fontAlign else return alignDefault end
end

function Emlib.scroll_createScroller(bkFrame, userColumns, scrollEffectiveProps) -- create FauxScrollFrame frame and lines
	local arrowWidth = 16 -- default
	if (not scrollEffectiveProps.defaultScroller) then arrowWidth = scrollEffectiveProps.scrollerWidth end
	local scrollerSuffix = "_scroller"
	---------------------------------------------- FauxScrollFrame frame
	--local fScroll = CreateFrame("ScrollFrame", bkFrame:GetName()..scrollerSuffix, bkFrame, "FauxScrollFrameTemplate")
	local fScroll = CreateFrame("ScrollFrame", bkFrame:GetName()..scrollerSuffix, bkFrame)
	--------------------------------- "FauxScrollFrameTemplate" template
--Emlib.setColoredFrame(fScroll, "red", nil, 0, 1)--/////
	fScroll:SetPoint("TOPLEFT", bkFrame.marginH, 0 - bkFrame.marginV)
	local tmpHeight = floor((bkFrame:GetHeight() - (bkFrame.marginV*2)) / bkFrame.lineHeight)
	local finalHeight = tmpHeight * bkFrame.lineHeight
	fScroll:SetSize(bkFrame:GetWidth() - (bkFrame.marginH * 2) - arrowWidth, finalHeight)
	fScroll.offsetPosUp = 0 -- init -- NEW 9.x perso
	fScroll:SetScript("OnLoad", function(self) ScrollFrame_OnLoad(self) end)
	fScroll:SetScript("OnVerticalScroll", function(self, offset)
		fScroll.offsetPosUp = offset -- NEW 9.x perso
		FauxScrollFrame_OnVerticalScroll(self, offset, bkFrame.lineHeight, Emlib.scrollUpdate(bkFrame))
	end)
	fScroll:SetScript("OnMouseWheel", function(self, delta) ScrollFrameTemplate_OnMouseWheel(self, delta) end)
	------------ scroll child
	local fChild = CreateFrame("Frame", "$parentScrollChildFrame", fScroll)
	fChild:SetSize(300, 334)
	fScroll:SetScrollChild(fChild)

	------------------------ slider : "ScrollBarTemplate" template
	--local fSlider = CreateFrame("Slider", "$parentScrollBar", fScroll, "ScrollBarTemplate")
	local fSlider = CreateFrame("Slider", "$parentScrollBar", fScroll)
--Emlib.setColoredFrame(fSlider, "blue", nil, 0, 1)--/////
	Emlib.setColoredFrame(fSlider, {bkFrame.background[1], bkFrame.background[2], bkFrame.background[3]}, nil, 0, scrollEffectiveProps.scrollerBarAlpha)
	fSlider:SetSize(arrowWidth, 0)
	fSlider:SetPoint("TOPLEFT", fScroll, "TOPRIGHT", bkFrame.marginH, -arrowWidth)
	fSlider:SetPoint("BOTTOMLEFT", fScroll, "BOTTOMRIGHT", bkFrame.marginH, arrowWidth)
	fSlider:SetOrientation("VERTICAL")
	--fSlider:SetScript("OnScrollRangeChanged", function() end)
	fSlider:SetScript("OnValueChanged", function(self, value)
		self:GetParent():SetVerticalScroll(value)
	end)
	
	------------ button (ascenseur)
	if (scrollEffectiveProps.defaultScroller) then
		fSlider.texture = fSlider:CreateTexture("$parentThumbTexture", "BACKGROUND")
		local pixScroller = "Interface\\Buttons\\UI-ScrollBar-Knob"
		fSlider.texture:SetTexture(pixScroller)
		fSlider.texture:SetSize(16, 24)					
		fSlider.texture:SetTexCoord(0.25, 0.75, 0.125, 0.875)
		fSlider:SetThumbTexture(fSlider.texture)
	else
		local pixScroller = fSlider:CreateTexture("ScrollerTex", "BACKGROUND")
		pixScroller:SetColorTexture(scrollEffectiveProps.scrollerColor[1], scrollEffectiveProps.scrollerColor[2], scrollEffectiveProps.scrollerColor[3], 0.9)
		pixScroller:SetAlpha(0.9) -- répétition pour bug SetColorTexture() ?
		pixScroller:SetAllPoints()
		pixScroller:SetSize(arrowWidth, 24)
		fSlider:SetThumbTexture(pixScroller)
	end
	
	------------ buttons (UP and DOWN arrows)
	local btnUp = Emlib.scroll_createScrollUpDownButtons("Up", fSlider, scrollEffectiveProps.defaultScroller, scrollEffectiveProps.scrollerWidth, scrollEffectiveProps.scrollerBarAlpha)
	local btnDown = Emlib.scroll_createScrollUpDownButtons("Down", fSlider, scrollEffectiveProps.defaultScroller, scrollEffectiveProps.scrollerWidth, scrollEffectiveProps.scrollerBarAlpha)
	--------------------------------- 

	bkFrame.slider = fSlider -- utility
	bkFrame.scroller = fScroll -- utility for update
	---------------------------------------------- frames in lines
	local checkedUserColumns = Emlib.scroll_setColumnsProps(userColumns, scrollEffectiveProps)
	Emlib.scroll_createLines(bkFrame, fScroll, checkedUserColumns)
	---------------------------------------------- for utility script
	bkFrame.column = {}
	for colIndex, thisColumnProps in ipairs(checkedUserColumns) do
		bkFrame.column[colIndex] = thisColumnProps
	end
end

function Emlib.scroll_createScrollUpDownButtons(kind, parent, defaultScroller, scrollerWidth, scrollerAlpha)
	local kind = kind or "Up"
	if (kind ~= "Up") then kind = "Down" end
	
	local btn = CreateFrame("Button", parent:GetName().."Scroll"..kind.."Button", parent)
	if (kind ~= "Up") then
		btn:SetPoint("TOP", parent, "BOTTOM", 0, 2)
		btn:SetScript("OnClick", function(self)
			parent:SetValue(parent:GetValue() + (parent:GetHeight() / 2))
			PlaySound(SOUNDKIT.U_CHAT_SCROLL_BUTTON)
		end)
	else
		btn:SetPoint("BOTTOM", parent, "TOP")
		btn:SetScript("OnClick", function(self)
			parent:SetValue(parent:GetValue() - (parent:GetHeight() / 2))
			PlaySound(SOUNDKIT.U_CHAT_SCROLL_BUTTON)
		end)
	end
	
	local pixBtn = Emlib.getStuffPath("Scroll"..kind..".tga", true)
								
	local pixBtnHi = Emlib.getStuffPath("Scroll"..kind..".tga", true)
	local pixBtnPushed = Emlib.getStuffPath("Scroll"..kind.."_hi.tga", true)
	if (defaultScroller) then
		scrollerAlpha = 1 -- override user settings
		-- bug texture highlight : XML = alphaMode="ADD"
		pixBtn = "Interface\\Buttons\\UI-ScrollBar-Scroll"..kind.."Button-Up"
		pixBtnHi = "Interface\\Buttons\\UI-ScrollBar-Scroll"..kind.."Button-Highlight"
		pixBtnPushed = "Interface\\Buttons\\UI-ScrollBar-Scroll"..kind.."Button-Down"
	end
	btn.texture = btn:CreateTexture("$parent_Tex", "BACKGROUND")
	btn.texture:SetTexture(pixBtn)
	btn.texture:SetAlpha(scrollerAlpha)
	btn.texture:SetAllPoints()
	if (defaultScroller) then
		btn:SetSize(18, 16)
		btn.texture:SetTexCoord(0.2, 0.8, 0.25, 0.75)
	else
		btn:SetSize(scrollerWidth, scrollerWidth)
	end
	
	btn.hitexture = btn:CreateTexture("$parent_TexHi", "BACKGROUND")
	btn.hitexture:SetTexture(pixBtnHi)
	btn.hitexture:SetAlpha(scrollerAlpha)
	btn.hitexture:SetAllPoints()
	btn:SetHighlightTexture(btn.hitexture)
	
	btn.pushedtexture = btn:CreateTexture("$parent_TexPushed", "BACKGROUND")
	btn.pushedtexture:SetTexture(pixBtnPushed)
	btn.pushedtexture:SetAlpha(scrollerAlpha)
	btn.pushedtexture:SetAllPoints()
	btn:SetPushedTexture(btn.pushedtexture)
	
	btn.disabledtexture = btn:CreateTexture("$parent_TexDisabled", "BACKGROUND")
	btn.disabledtexture:SetTexture(pixBtn)
	btn.disabledtexture:SetAlpha(scrollerAlpha)
	btn.disabledtexture:SetAllPoints()
	btn:SetDisabledTexture(btn.disabledtexture)
	
	return btn
end

function Emlib.scroll_createLines(bkFrame, fScroll, checkedUserColumns) -- lines
	local scrollLineNamePrefix = bkFrame:GetName().."_line_"
	
	bkFrame.columns = #checkedUserColumns -- utility
	bkFrame.line = {} -- utility for update
	bkFrame.bkLine = {} -- utility for update
	local tblColumnsFramesName -- table of all the frames' name in a line
	
	local frameLine
	bkFrame.maxFrameLines = ceil(bkFrame.maxHeight / bkFrame.lineHeight) -- number of line frames created
	for lineIndex = 1, bkFrame.maxFrameLines do
		------------------ line frame (a background for each line)
		frameLine = CreateFrame("Button", scrollLineNamePrefix..lineIndex, bkFrame)
		frameLine:SetSize(fScroll:GetWidth(), bkFrame.lineHeight - bkFrame.lineSpacing)
		frameLine:SetPoint("TOPLEFT", fScroll, "TOPLEFT", 0, 0 - ((lineIndex - 1) * bkFrame.lineHeight))
		
		if (bkFrame.mouseOverLineColor ~= nil) then -- for color change on mouseOver the lines
			Emlib.setColoredButton(frameLine, bkFrame.background, nil, 0, bkFrame.alpha, bkFrame.mouseOverLineColor)
		end
		------------------ sub-frames in each line (depending on data)
		bkFrame.line[lineIndex] = {} -- utility for update
		Emlib.scroll_createColumns(bkFrame, frameLine, checkedUserColumns, lineIndex, bkFrame.cellSpacing)
		------------------ 
		frameLine.columnsName = tblColumnsFramesName -- handler for utilities: list of all sub-frames in this line
		frameLine.lineIndex = lineIndex -- handler for utilities
		bkFrame.bkLine[lineIndex] = frameLine -- utility for update
		frameLine:Hide() -- will be shown on update
	end
end

function Emlib.scroll_setColumnsProps(userColumns, scrollEffectiveProps) -- check user colums properties (and add missing ones)
	local defaultColumnProps = Emlib.scroll_getColumnDefaultProperties()
	
	local userCheckedColumns = {}
	for colIndex, column in ipairs(userColumns) do
		local tbl = {}
		tbl["name"] = colIndex
		
		if (column["width"] == nil) then
			tbl["width"] = 0
		else
			tbl["width"] = Emlib.scroll_checkColumnProp(column, defaultColumnProps, scrollEffectiveProps, "width", nil) -- no prop type (% admitted)
		end
		if (column["height"] == nil) then
			tbl["height"] = 0
		else
			tbl["height"] = Emlib.scroll_checkColumnProp(column, defaultColumnProps, scrollEffectiveProps, "height", "number")
		end
		
		tbl["type"] = column["type"] or ""
		tbl["multilines"] = Emlib.isTrue(column["multilines"])
		
		tbl["fontAlpha"] = Emlib.scroll_checkColumnProp(column, defaultColumnProps, scrollEffectiveProps, "fontAlpha", "number")
		tbl["font"] = Emlib.scroll_checkColumnProp(column, defaultColumnProps, scrollEffectiveProps, "font", "string")
		tbl["fontSize"] = Emlib.scroll_checkColumnProp(column, defaultColumnProps, scrollEffectiveProps, "fontSize", "number")
		tbl["fontOutline"] = Emlib.scroll_checkColumnProp(column, defaultColumnProps, scrollEffectiveProps, "fontOutline", "boolean")
		tbl["padding"] = Emlib.scroll_checkColumnProp(column, defaultColumnProps, scrollEffectiveProps, "padding", "number")
		if (tonumber(tbl["padding"]) == nil) then tbl["padding"] = tbl["fontSize"]/3 end
		
		local fontColor = Emlib.scroll_checkColumnProp(column, defaultColumnProps, scrollEffectiveProps, "fontColor", nil)
		tbl["fontColor"] = Emlib.getColor(fontColor, "table")
		
		local fontAlign = strupper(Emlib.scroll_checkColumnProp(column, defaultColumnProps, scrollEffectiveProps, "fontAlign", "string"))
		tbl["fontAlign"]= Emlib.scroll_getTextAlign(fontAlign, "LEFT")
		
		tbl["background"] = Emlib.scroll_checkColumnProp(column, defaultColumnProps, scrollEffectiveProps, "background", nil)
		tbl["alpha"] = Emlib.scroll_checkColumnProp(column, defaultColumnProps, scrollEffectiveProps, "alpha", "number")
		tbl["border"] = Emlib.scroll_checkColumnProp(column, defaultColumnProps, scrollEffectiveProps, "border", "number")
		tbl["borderColor"] = Emlib.scroll_checkColumnProp(column, defaultColumnProps, scrollEffectiveProps, "borderColor", nil)
		tbl["mouseOverLineColor"] = Emlib.scroll_checkColumnProp(column, defaultColumnProps, scrollEffectiveProps, "mouseOverLineColor", nil)
		
		tbl["template"] = Emlib.scroll_checkColumnProp(column, defaultColumnProps, scrollEffectiveProps, "template", nil)
		tbl["templateType"] = Emlib.scroll_checkColumnProp(column, defaultColumnProps, scrollEffectiveProps, "templateType", nil)
		
		tbl["script"] = column["script"] or ""
		
		userCheckedColumns[colIndex] = tbl
	end
	return userCheckedColumns
end

------------------------------------------------------------------- SUB-FRAMES (COLUMNS) CREATION
function Emlib.scroll_createColumns(bkFrame, frameLine, userColumns, lineIndex, cellSpacing)
	-- creates columns for a passed line frame
	bkFrame.line[lineIndex].column = {} -- utility for update
	
	local columnFrameName
	local prevColumnFrame -- for positioning
	local filledSpace = 0 -- space already filled with defined column/frames
	for colIndex, thisColumnProps in ipairs(userColumns) do
		local columnType = "Frame"
		if (strupper(thisColumnProps.type) == "BUTTON") then
			columnType = "Button"
		elseif (strupper(thisColumnProps.type) == "EDITBOX") then
			columnType = "EditBox"
		end
		if (thisColumnProps.template == "") then
			thisColumnProps.template = nil -- secu
			thisColumnProps.templateType = nil -- secu
		else
			if (thisColumnProps.templateType ~= nil) then columnType = thisColumnProps.templateType or "Frame" end
		end
		columnFrameName = frameLine:GetName().."_"..thisColumnProps.name
		local frame = CreateFrame(columnType, columnFrameName, frameLine, thisColumnProps.template)
		---------------------------------- size
		local columnHeight = frameLine:GetHeight() -- default
		if (thisColumnProps.height > 0) then columnHeight = max(6, min(columnHeight, frameLine:GetHeight())) end
		local columnWidth = thisColumnProps.width
		if (strfind(columnWidth, "%", 1 , true) ~= nil) then
			---------------- % value
			columnWidth = string.gsub(columnWidth, "%%", "")
			columnWidth = tonumber(columnWidth) or 1 -- 1% minimum
			columnWidth = frameLine:GetWidth() * columnWidth / 100
		else
			---------------- px value
			columnWidth = tonumber(columnWidth) or 0 -- if 0 : remaining space
			if (columnWidth < 1) then columnWidth = frameLine:GetWidth() - filledSpace end
		end
		columnWidth = math.max(4, columnWidth) -- 4 to make column visible if not enough space
		filledSpace = filledSpace + columnWidth
		if (colIndex == #userColumns) then
			frame:SetSize(columnWidth, columnHeight) -- no cellSpacing for the last column
		else
			frame:SetSize(columnWidth - cellSpacing, columnHeight)
		end
		---------------------------------- position
		if (colIndex > 1) then
			frame:SetPoint("LEFT", prevColumnFrame, "RIGHT", cellSpacing, 0)
		else
			frame:SetPoint("LEFT", frameLine, "LEFT", 0, 0)
		end
		---------------------------------- text (ou contenu EditBox)
		local fontSize = thisColumnProps.fontSize -- shortcut
		if (columnType == "EditBox") then
			frame:SetMultiLine(thisColumnProps.multilines)
			frame:SetAutoFocus(false)
			frame:SetHighlightColor(0.82,0.79,0.05) -- couleur sélection texte
			frame:SetTextInsets(8, 8, -1, 0) -- left, right, top, bottom
			frame:SetFontObject("ChatFontNormal")
			frame:SetFont(thisColumnProps.font, fontSize, "NONE")
			frame:SetJustifyV("TOP") -- ou "BOTTOM", or "MIDDLE"
			frame:SetText("")
		else
			frame.text = frame:CreateFontString(columnFrameName.."_Text", "OVERLAY", "GameFontNormalSmall")
			Emlib.setFontStyle(frame.text, thisColumnProps.font, fontSize, thisColumnProps.fontColor, thisColumnProps.fontOutline, thisColumnProps.fontAlpha)
			frame.text:SetWordWrap(false)
			frame.text:SetJustifyH(thisColumnProps.fontAlign)
			--frame.text:SetJustifyV("TOP") -- ou "BOTTOM", or "MIDDLE"
			
			frame.text:SetWidth(ceil(frame:GetWidth()) - (thisColumnProps.padding * 2))
			if (thisColumnProps.fontAlign == "LEFT") then
				frame.text:SetPoint("LEFT", thisColumnProps.padding, fontSize/12)
			elseif (thisColumnProps.fontAlign == "RIGHT") then
				frame.text:SetPoint("RIGHT", 0 - thisColumnProps.padding, fontSize/12)
			else
				frame.text:SetPoint("CENTER", 0, fontSize/12)
			end
			frame.text:SetNonSpaceWrap(false)
		end
		---------------------------------- background
		local background = thisColumnProps.background -- shortcut
		
		if (type(background) == "table") then
			--------------- color table
			Emlib.setColoredRegion(frame, background, thisColumnProps.borderColor, thisColumnProps.border, thisColumnProps.alpha, thisColumnProps.mouseOverLineColor)
		else
			if ((string.find(background, "/", 1, true) ~= nil) or (string.find(background, "_", 1, true) ~= nil)) then
				--------------- texture name
				Emlib.setTexturedFrame(frame, "$parent_tex", background, columnWidth, frame:GetHeight(), thisColumnProps.alpha, thisColumnProps.mouseOverLineColor)
			else
				--------------- color string
				Emlib.setColoredRegion(frame, background, thisColumnProps.borderColor, thisColumnProps.border, thisColumnProps.alpha, thisColumnProps.mouseOverLineColor)
			end
		end
		---------------------------------- script
		if (not Emlib.isEmpty(thisColumnProps.script)) then
			if (columnType == "Button") then
	--print(thisColumnProps.script.." ("..Emlib.kwote(frame:GetName())..","..Emlib.kwote(mouse)..","..lineIndex..")")--/////
				frame:RegisterForClicks("LeftButtonUp","RightButtonUp") -- nécessaire pour le clic Droit
				frame:SetScript("OnClick", function(self, mouse)
					RunScript(thisColumnProps.script.."("..Emlib.kwote(bkFrame:GetName())..","..Emlib.kwote(self:GetName())..","..Emlib.kwote(mouse)..","..lineIndex..")")
				end)
			elseif (columnType == "EditBox") then
				frame:SetScript("OnEscapePressed", function(self) -- deselect
					self:HighlightText(0, 0)
					self:ClearFocus()
				end)
				frame:SetScript("OnEnterPressed", function(self)
					RunScript(thisColumnProps.script.."("..Emlib.kwote(bkFrame:GetName())..","..Emlib.kwote(self:GetName())..", nil ,"..lineIndex..")")
				end)
			end
		end
		---------------------------------- internal
		prevColumnFrame = frame -- for positioning
		frame.lineIndex = lineIndex
		frame.columnIndex = colIndex
		
		bkFrame.line[lineIndex].column[colIndex] = frame -- utility for update
	end
end
